// modules import
const ffi = require('ffi-napi');

const ref = require('ref-napi');

const bodyParser = require('body-parser');

const express = require('express');

// DLL & API functions import
var oldPath = process.env.PATH;
// path to uFCoder library to load
var dllPath = "ufr-lib/windows/x86_64/uFCoder-x86_64.dll"

// necessary load for OriginalityCheck to work properly
var libbSSLPath = "ufr-lib/windows/x86_64/libeay32.dll"
var libSSL = ffi.Library(libbSSLPath)


var uFCoder = ffi.Library(dllPath, {
  "ReaderOpen": ['int', []],
  "ReaderOpenEx": ['int', ['int', 'string', 'int', 'string']],
  "ReaderClose": ['int', []],
  "GetCardIdEx": ['int', ['byte *', 'byte *', 'byte *']],
  "LinearRead": ['int', ['byte*', 'ushort', 'ushort', 'ushort*', 'byte', 'byte']],
  "LinearRead_PK": ['int', ['byte*', 'ushort', 'ushort', 'ushort*', 'byte', 'byte *']],
  "LinearWrite": ['int', ['byte*', 'ushort', 'ushort', 'ushort*', 'byte', 'byte']],
  "LinearWrite_PK": ['int', ['byte*', 'ushort', 'ushort', 'ushort*', 'byte', 'byte *']],
  "LinearFormatCard": ['int', ['byte*', 'byte', 'byte', 'byte', 'byte*', 'byte*', 'byte', 'byte']],
  "GetCardSize": ['int', ['int*', 'int*']],
  "GetDlogicCardType": ['int', ['byte*']],
  "ReaderUISignal": ['int', ['int', 'int']],
  "UFR_Status2String": ['string', ['int']],
  "GetDllVersionStr": ['string', []],

  "nt4h_set_global_parameters": ['int', ['byte', 'byte', 'byte']],
  "nt4h_change_standard_file_settings_pk": ['int', ['byte *', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte']],
  "nt4h_change_standard_file_settings": ['int', ['byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte']],
  "nt4h_change_sdm_file_settings_pk": ['int', ['byte *', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'int', 'int', 'int', 'int', 'int', 'int', 'int', 'int']],
  "nt4h_change_sdm_file_settings": ['int', ['byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'int', 'int', 'int', 'int', 'int', 'int', 'int', 'int']],
  "nt4h_get_file_settings": ['int', ['byte', 'byte *', 'byte *', 'byte *', 'int *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'int *', 'int *', 'int *', 'int *', 'int *', 'int *', 'int *', 'int *']],

  "nt4h_tt_get_file_settings": ['int', ['byte', 'byte *', 'byte *', 'byte *', 'int *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'byte *', 'int *', 'int *', 'int *', 'int *', 'int *', 'int *', 'int *', 'int *', 'byte *', 'int *']],
  "nt4h_set_rid_pk": ['int', ['byte *']],
  "nt4h_set_rid": ['int', ['byte']],
  "nt4h_get_uid_pk": ['int', ['byte *', 'byte', 'byte *']],
  "nt4h_get_uid": ['int', ['byte', 'byte', 'byte *']],
  "nt4h_change_key_pk": ['int', ['byte *', 'byte', 'byte *', 'byte *']],
  "nt4h_change_key": ['int', ['byte', 'byte', 'byte *', 'byte *']],
  "nt4h_get_sdm_ctr_pk": ['int', ['byte *', 'byte', 'byte', 'int *']],

  "nt4h_get_sdm_ctr": ['int', ['byte', 'byte', 'byte', 'int *']],
  "nt4h_get_sdm_ctr_no_auth": ['int', ['byte', 'int *']],
  "nt4h_check_sdm_mac": ['int', ['int', 'byte *', 'byte *', 'byte *', 'byte', 'byte *']],
  "nt4h_decrypt_sdm_enc_file_data": ['int', ['int', 'byte *', 'byte *', 'byte *', 'byte']],
  "nt4h_decrypt_picc_data": ['int', ['byte *', 'byte *', 'byte *', 'byte *', 'int *']],
  "nt4h_tt_change_sdm_file_settings_pk": ['int', ['byte *', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'int', 'int', 'int', 'int', 'int', 'int', 'int', 'int', 'byte', 'int']],
  "nt4h_tt_change_sdm_file_settings": ['int', ['byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'byte', 'int', 'int', 'int', 'int', 'int', 'int', 'int', 'int', 'byte', 'int']],
  "nt4h_rid_read_ecc_signature_pk": ['int', ['byte *', 'byte', 'byte *', 'byte *', 'byte *']],

  "nt4h_rid_read_ecc_signature": ['int', ['byte', 'byte', 'byte *', 'byte *', 'byte *']],
  "nt4h_get_tt_status_pk": ['int', ['byte *', 'byte', 'byte *', 'byte *']],
  "nt4h_get_tt_status": ['int', ['byte', 'byte', 'byte *', 'byte *']],
  "nt4h_get_tt_status_no_auth": ['int', ['byte *', 'byte *']],
  "nt4h_enable_tt_pk": ['int', ['byte *', 'byte']],
  "nt4h_enable_tt": ['int', ['byte', 'byte']],
  
  "ReadECCSignatureExt": ['int', ['byte *', 'byte *', 'byte *', 'byte *', 'byte *']],
  "OriginalityCheck": ['int', ['byte *', 'byte *', 'byte', 'byte']],


  "uFR_int_DesfireWriteAesKey": ['int', ['byte', 'byte *']],
  "SetISO14443_4_Mode": ['int', []],
  "APDUPlainTransceive": ['int', ['byte *', 'int', 'byte *', 'int *']],
  "s_block_deselect": ['int', ['byte']]

});

process.env['PATH'] = oldPath;

console.log(uFCoder.GetDllVersionStr())

// T4T authentication constants
const T4T_AUTHENTICATION = {
  T4T_WITHOUT_PWD_AUTH: 0x60,
  T4T_PK_PWD_AUTH: 0x80,
  T4T_RKA_PWD_AUTH: 0x02,
};

// Express setup

const hostname = '127.0.0.1';

const port = 3000;

var app = express();

app.use(bodyParser.json());

app.use(express.static(__dirname + '/public'));

////////////////////////////////////////////////////////////////////////////////////////
// uFR Reader Functions
////////////////////////////////////////////////////////////////////////////////////////

app.post("/readerOpenEx", function (req, res) {

  console.log(req.body);

  var reader_type = req.body.ReaderType;

  var port_name = req.body.PortName;
  if (port_name == '0') {
    port_name = '';
  }

  var port_interface = req.body.PortInterface;
  if (port_interface == 3) // || port_interface == "U")
  {
    port_interface = 85; // UDP //port_interface.charCodeAt(0);
  } else if (port_interface == 4) {
    port_interface = 84; // TCP
  } else {
    port_interface = (port_interface);
  }

  var arg = req.body.Arg;
  if (arg == '0') {
    arg = '';
  }

  var status = uFCoder.ReaderOpenEx(reader_type, port_name, port_interface, arg);
  var success = "";
  if (status == 0) {
    uFCoder.ReaderUISignal(1, 1);
    success = "Function status: ReaderOpenEx() - successful"
  } else {
    success = "Function status: ReaderOpenEx() - failed"
  }

  var response = {
    "Status": uFR_NT4H_Status2String(status),
    "Success": success
  }

  //console.log("Reader open response: ");
  console.log(response);

  res.send(JSON.stringify(response));
});

////////////////////////////////////////////////////////////////////////////////////////

app.post("/readerOpen", function (req, res) {

  var status = uFCoder.ReaderOpen();
  var success = "";
  if (status == 0) {
    uFCoder.ReaderUISignal(1, 1);
    success = "Function status: ReaderOpen() - successful"
  } else {
    success = "Function status: ReaderOpen() - failed"
  }
  var response = {
    "Status": uFR_NT4H_Status2String(status),
    "Success": success
  }
  res.send(JSON.stringify(response));
});

////////////////////////////////////////////////////////////////////////////////////////

app.post("/readerClose", function (req, res) {

  var status = uFCoder.ReaderClose();
  var success = "";
  if (status == 0) {
    success = "Function status: ReaderClose() - successful"
  } else {
    success = "Function status: ReaderClose() - failed"
  }
  var response = {
    "Status": uFR_NT4H_Status2String(status),
    "Success": success
  }
  res.send(JSON.stringify(response));
});

///////////////////////////////////////////////////////////////////////
// NT4H Functions
///////////////////////////////////////////////////////////////////////

app.post("/getFileSettings", function (req, res) {
  console.log("GET FILE SETTINGS!");
  console.log(req.body);

  var response = {
    "Status": "",
    "UfrStatus": "",
    "Success": "",

    "file_type": "",
    "sdm_enable": "",
    "file_size": "",
    "comm_mode": "",
    "read_key_no": "",
    "write_key_no": "",
    "read_write_key_no": "",
    "change_key_no": "",
    "uid_enable": "",
    "read_ctr_enable": "",
    "read_ctr_limit_enable": "",
    "enc_file_data_enable": "",
    "meta_data_key_no": "",
    "file_data_read_key_no": "",
    "read_ctr_key_no": "",
    "uid_offset": "",
    "read_ctr_offset": "",
    "picc_data_offset": "",
    "mac_input_offset": "",
    "enc_offset": "",
    "enc_length": "",
    "mac_offset": "",
    "read_ctr_limit": "",
    "tt_status_enable": "",
    "tt_status_offset": "",
  }

  var file_no = ref.alloc('byte', Number(req.body.file_no))

  var file_type = ref.alloc('byte', 0);
  var sdm_enable = ref.alloc('byte', 0);
  var file_size = ref.alloc('int', 0);

  var comm_mode = ref.alloc('byte', 0);

  var read_key_no = ref.alloc('byte', 0);
  var write_key_no = ref.alloc('byte', 0);
  var read_write_key_no = ref.alloc('byte', 0);
  var change_key_no = ref.alloc('byte', 0);

  var uid_enable = ref.alloc('byte', 0);
  var read_ctr_enable = ref.alloc('byte', 0);
  var read_ctr_limit_enable = ref.alloc('byte', 0);
  var enc_file_data_enable = ref.alloc('byte', 0);

  var meta_data_key_no = ref.alloc('byte', 0);
  var file_data_read_key_no = ref.alloc('byte', 0);
  var read_ctr_key_no = ref.alloc('byte', 0);

  var uid_offset = ref.alloc('int', 0);
  var read_ctr_offset = ref.alloc('int', 0);
  var picc_data_offset = ref.alloc('int', 0);
  var mac_input_offset = ref.alloc('int', 0);

  var enc_offset = ref.alloc('int', 0);
  var enc_length = ref.alloc('int', 0);
  var mac_offset = ref.alloc('int', 0);
  var read_ctr_limit = ref.alloc('int', 0);

  var tt_status_enable = ref.alloc('byte', 0);
  var tt_status_offset = ref.alloc('byte', 0);

  var status = 0x1001;

  var dl_card_type = ref.alloc('byte', 0);
  status = uFCoder.GetDlogicCardType(dl_card_type);
  if (status) {
    response.Success = "Function status: Geting card type failed."
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
    //res.send(JSON.stringify(response));
    console.log(response)
    uFCoder.s_block_deselect(100)
    res.send(JSON.stringify(response));
    return;
  }

  if (dl_card_type.deref() == 0x13) // DL_NTAG_424_DNA_TT
  {
    //nt4h_tt_get_file_settings
    status = uFCoder.nt4h_tt_get_file_settings(file_no.deref(), file_type, comm_mode, sdm_enable, file_size,
      read_key_no, write_key_no, read_write_key_no, change_key_no,
      uid_enable, read_ctr_enable, read_ctr_limit_enable, enc_file_data_enable,
      meta_data_key_no, file_data_read_key_no, read_ctr_key_no,
      uid_offset, read_ctr_offset, picc_data_offset,
      mac_input_offset, enc_offset, enc_length, mac_offset, read_ctr_limit,
      tt_status_enable, tt_status_offset);

  } else {
    //nt4h_get_file_settings
    status = uFCoder.nt4h_get_file_settings(file_no.deref(), file_type, comm_mode, sdm_enable, file_size,
      read_key_no, write_key_no, read_write_key_no, change_key_no,
      uid_enable, read_ctr_enable, read_ctr_limit_enable, enc_file_data_enable,
      meta_data_key_no, file_data_read_key_no, read_ctr_key_no,
      uid_offset, read_ctr_offset, picc_data_offset,
      mac_input_offset, enc_offset, enc_length, mac_offset, read_ctr_limit);

  }

  if (status) {
    response.Success = "Function status: Failed in getting file settings.";
  } else {
    response.Success = "Function status: Getting file settings was successful.";
  
    response.file_type = file_type.deref();
    
    response.file_size = file_size.deref();
    response.comm_mode = comm_mode.deref();
    response.read_key_no = read_key_no.deref();
    response.write_key_no = write_key_no.deref();
    response.read_write_key_no = read_write_key_no.deref();
    response.change_key_no = change_key_no.deref();

    response.sdm_enable = sdm_enable.deref();
    response.uid_enable = uid_enable.deref();
    response.read_ctr_enable = read_ctr_enable.deref();
    response.read_ctr_limit_enable = read_ctr_limit_enable.deref();
    response.enc_file_data_enable = enc_file_data_enable.deref();

    response.meta_data_key_no = meta_data_key_no.deref();
    response.file_data_read_key_no = file_data_read_key_no.deref();
    response.read_ctr_key_no = read_ctr_key_no.deref();

    response.uid_offset = uid_offset.deref();
    response.read_ctr_offset = read_ctr_offset.deref();
    response.picc_data_offset = picc_data_offset.deref();
    response.mac_input_offset = mac_input_offset.deref();
    response.enc_offset = enc_offset.deref();
    response.enc_length = enc_length.deref();
    response.mac_offset = mac_offset.deref();
    response.read_ctr_limit = read_ctr_limit.deref();

    response.tt_status_enable = tt_status_enable.deref();
    response.tt_status_offset = tt_status_offset.deref();
    
  }
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
     //res.send(JSON.stringify(response));
    console.log(response)

  res.send(JSON.stringify(response));
});

///////////////////////////////////////////////////////////////////////

app.post("/setFileSettings", function (req, res) {
  console.log("GET FILE SETTINGS!");
  console.log(req.body);

  var response = {
    "Status": "",
    "UfrStatus": "",
    "Success": "",
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var file_no = ref.alloc('byte', Number(req.body.file_no))

  var sdm_enable = ref.alloc('byte', Number(req.body.sdm_enable));

  var curr_comm_mode = ref.alloc('byte', 3);
  var new_comm_mode = ref.alloc('byte', Number(req.body.new_comm_mode));

  var curr_change_key_no = ref.alloc('byte', Number(req.body.change_key_no));
  var read_key_no = ref.alloc('byte', Number(req.body.read_key_no));
  var write_key_no = ref.alloc('byte', Number(req.body.write_key_no));
  var read_write_key_no = ref.alloc('byte', Number(req.body.read_write_key_no));
  var new_change_key_no = ref.alloc('byte', Number(req.body.new_change_key_no));

  var uid_enable = ref.alloc('byte', Number(req.body.uid_enable));
  var read_ctr_enable = ref.alloc('byte', Number(req.body.read_ctr_enable));
  var read_ctr_limit_enable = ref.alloc('byte', Number(req.body.read_ctr_limit_enable));
  var enc_file_data_enable = ref.alloc('byte', Number(req.body.enc_file_data_enable));

  var meta_data_key_no = ref.alloc('byte', Number(req.body.meta_data_key_no));
  var file_data_read_key_no = ref.alloc('byte', Number(req.body.file_data_read_key_no));
  var read_ctr_key_no = ref.alloc('byte', Number(req.body.read_ctr_key_no));

  var uid_offset = ref.alloc('int', Number(req.body.uid_offset));
  var read_ctr_offset = ref.alloc('int', Number(req.body.read_ctr_offset));
  var picc_data_offset = ref.alloc('int', Number(req.body.picc_data_offset));
  var mac_input_offset = ref.alloc('int', Number(req.body.mac_input_offset));

  var enc_offset = ref.alloc('int', Number(req.body.enc_offset));
  var enc_length = ref.alloc('int', Number(req.body.enc_length));
  var mac_offset = ref.alloc('int', Number(req.body.mac_offset));
  var read_ctr_limit = ref.alloc('int', Number(req.body.read_ctr_limit));

  var tt_status_enable = ref.alloc('byte', Number(req.body.tt_status_enable));
  var tt_status_offset = ref.alloc('byte', Number(req.body.tt_status_offset));

  var status = 0x1001;

  var dl_card_type = ref.alloc('byte', 0);
  status = uFCoder.GetDlogicCardType(dl_card_type);
  if (status) {
    response.Success = "Function status: Geting card type failed."
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
    //res.send(JSON.stringify(response));
    console.log(response)
    uFCoder.s_block_deselect(100)
    res.send(JSON.stringify(response));
    return;
  }

  if (sdm_enable.deref() == 0) // sdm not enabled
  {
    if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_RKA_PWD_AUTH)
    {
      //RK change
      status = uFCoder.nt4h_change_standard_file_settings(readerKeyIndex.deref(), file_no.deref(), curr_change_key_no.deref(),
      curr_comm_mode.deref(), new_comm_mode.deref(), read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref());
    } else {
      //PK change
      status = uFCoder.nt4h_change_standard_file_settings_pk(providedKey, file_no.deref(), curr_change_key_no.deref(),
      curr_comm_mode.deref(), new_comm_mode.deref(), read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref());
    }
  } else {
    // change SDM
    if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_RKA_PWD_AUTH) {
      // reader key authentication
      if (dl_card_type.deref() == 0x13) // if DL_NTAG_424_DNA_TT
      {
        status = uFCoder.nt4h_tt_change_sdm_file_settings(readerKeyIndex.deref(), file_no.deref(), curr_change_key_no.deref(),
          curr_comm_mode.deref(), new_comm_mode.deref(),
          read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
          uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
          meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
          uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
          enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref(),
          tt_status_enable.deref(), tt_status_offset.deref())

      } else {
        //status = uFCoder.nt4h_change_sdm_file_settings()
        status = uFCoder.nt4h_change_sdm_file_settings(readerKeyIndex.deref(), file_no.deref(), curr_change_key_no.deref(),
          curr_comm_mode.deref(), new_comm_mode.deref(),
          read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
          uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
          meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
          uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
          enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref())
      }
    } else {
      if (dl_card_type.deref() == 0x13) // if DL_NTAG_424_DNA_TT
      {
        status = uFCoder.nt4h_tt_change_sdm_file_settings_pk(providedKey, file_no.deref(), curr_change_key_no.deref(),
          curr_comm_mode.deref(), new_comm_mode.deref(),
          read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
          uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
          meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
          uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
          enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref(),
          tt_status_enable.deref(), tt_status_offset.deref())
      } else {
        status = uFCoder.nt4h_change_sdm_file_settings_pk(providedKey, file_no.deref(), curr_change_key_no.deref(),
          curr_comm_mode.deref(), new_comm_mode.deref(),
          read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
          uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
          meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
          uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
          enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref())
      }
    }
  }

  if (status) {
    response.Success = "Function status: Failed in setting file settings.";
  } else {
    response.Success = "Function status: Setting file settings was successful.";
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status
  //res.send(JSON.stringify(response));
  console.log(response)

res.send(JSON.stringify(response));

});

///////////////////////////////////////////////////////////////////////

app.post("/readSDMData", function (req, res) {

  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0,
    "ndefHexDataStr": "",
    "ndefDataStr": "",
    "piccDataStr": "",
    "piccUidStr": "",
    "piccSDMReadingCounter": "",
    "piccEncryptedFileData": "",
    "piccFileData": "",
    "piccASCIIMACData":"",
    "piccASCIIMACInputData": ""
  }

  var global_file_no = ref.alloc('byte', 2);
  var global_key_no = ref.alloc('byte', 14); // NDEF for SDM must be set as 'Read Access - 14'
  var global_comm_mode = ref.alloc('byte', 0);

  var meta_data_key_pk = prepareAesKey(req.body.metaAESKeyStr);
  var file_data_read_key_pk = prepareAesKey(req.body.fileAESKey);

  var file_no = ref.alloc('byte', Number(req.body.file_no))

  var file_type = ref.alloc('byte', 0);
  var sdm_enable = ref.alloc('byte', 0);
  var file_size = ref.alloc('int', 0);

  var comm_mode = ref.alloc('byte', 0);

  var read_key_no = ref.alloc('byte', 0);
  var write_key_no = ref.alloc('byte', 0);
  var read_write_key_no = ref.alloc('byte', 0);
  var change_key_no = ref.alloc('byte', 0);

  var uid_enable = ref.alloc('byte', 0);
  var read_ctr_enable = ref.alloc('byte', 0);
  var read_ctr_limit_enable = ref.alloc('byte', 0);
  var enc_file_data_enable = ref.alloc('byte', 0);

  var meta_data_key_no = ref.alloc('byte', 0);
  var file_data_read_key_no = ref.alloc('byte', 0);
  var read_ctr_key_no = ref.alloc('byte', 0);

  var uid_offset = ref.alloc('int', 0);
  var read_ctr_offset = ref.alloc('int', 0);
  var picc_data_offset = ref.alloc('int', 0);
  var mac_input_offset = ref.alloc('int', 0);

  var enc_offset = ref.alloc('int', 0);
  var enc_length = ref.alloc('int', 0);
  var mac_offset = ref.alloc('int', 0);
  var read_ctr_limit = ref.alloc('int', 0);

  var tt_status_enable = ref.alloc('byte', 0);
  var tt_status_offset = ref.alloc('byte', 0);

  var status = 0x1001;

  var dl_card_type = ref.alloc('byte', 0);
  status = uFCoder.GetDlogicCardType(dl_card_type);
  if (status) {
    response.Success = "Function status: Geting card type failed."
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
    //res.send(JSON.stringify(response));
    console.log(response)
    uFCoder.s_block_deselect(100)

    res.send(JSON.stringify(response));
    return;
  }

  if (dl_card_type.deref() == 0x13) // DL_NTAG_424_DNA_TT
  {
    //nt4h_tt_get_file_settings
    status = uFCoder.nt4h_tt_get_file_settings(global_file_no.deref(), file_type, comm_mode, sdm_enable, file_size,
      read_key_no, write_key_no, read_write_key_no, change_key_no,
      uid_enable, read_ctr_enable, read_ctr_limit_enable, enc_file_data_enable,
      meta_data_key_no, file_data_read_key_no, read_ctr_key_no,
      uid_offset, read_ctr_offset, picc_data_offset,
      mac_input_offset, enc_offset, enc_length, mac_offset, read_ctr_limit,
      tt_status_enable, tt_status_offset);

  } else {
    //nt4h_get_file_settings
    status = uFCoder.nt4h_get_file_settings(global_file_no.deref(), file_type, comm_mode, sdm_enable, file_size,
      read_key_no, write_key_no, read_write_key_no, change_key_no,
      uid_enable, read_ctr_enable, read_ctr_limit_enable, enc_file_data_enable,
      meta_data_key_no, file_data_read_key_no, read_ctr_key_no,
      uid_offset, read_ctr_offset, picc_data_offset,
      mac_input_offset, enc_offset, enc_length, mac_offset, read_ctr_limit);

  }

  if (status) {
    response.Success = "Function status: Failed in getting file settings.";
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
    console.log(response)
    uFCoder.s_block_deselect(100)

    res.send(JSON.stringify(response));
    return;
  }
  var sdm_mode = (sdm_enable.deref() == 1) || (communication_mode.deref() == 0) && (read_key_no.deref() == 0x0E)
  if (sdm_mode == false)
  {
    response.Success = "Function status: File is not in SDM mode.";
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
    //res.send(JSON.stringify(response));
    console.log(response)
    uFCoder.s_block_deselect(100)

    res.send(JSON.stringify(response));
  }

  var global_file_no = ref.alloc('byte', 2);
  var global_key_no = ref.alloc('byte', 0x0E); // ndef read key set
  var global_comm_mode = ref.alloc('byte', 0); // set to plain for ndef write

  status = uFCoder.nt4h_set_global_parameters(global_file_no.deref(), global_key_no.deref(), global_comm_mode.deref());
  if (status) {
    response.Success = "Function status: Global parameters setting failed"
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status;

    console.log(response)
    res.send(JSON.stringify(response));
    return;
  }

  var read_data = Buffer.alloc(file_size.deref(), 0)

  var bytes_ret = ref.alloc('int', 0)
  
  status = uFCoder.LinearRead(read_data, 0, file_size.deref(), bytes_ret, T4T_AUTHENTICATION.T4T_WITHOUT_PWD_AUTH, 0);
  if (status) {
    response.Success = "Function status: Read NDEF data failed.";
    response.Status = uFR_NT4H_Status2String(status)
    console.log(response)

    res.send(JSON.stringify(response));
    return;
  }


  //read_data


  //var read_data = rapdu.slice(0, -2)
  var read_data_len = read_data.length

  var read_data_str = "";
  for (i = 0; i < read_data_len; i++) {
    read_data_str += read_data[i].toString(16).padStart(2, '0').toUpperCase() + ":";
  }

  read_data_str = read_data_str.slice(0, -1).replace(/\0/g, '');

  var ndef_length = read_data[4];
  //var ndef_le = rapdu.readInt16LE(4);
  var ndef_context_buffer = read_data.slice(7).slice(0, (ndef_length - 1))
  // skipping NDEF header and slicing NDEF length to get the data

  var ndef_context_dataStr = ndef_context_buffer.toString('utf-8');

  // DECRYPT PICC DATA

  var sdm_read_cnt = ref.alloc('int', 0);
  var picc_uid = Buffer.alloc(7);

  if (meta_data_key_no.deref() <=4)
  {
    // PICC DATA EXIST - meta data key != 0x0F
    var picc_data = read_data.slice(picc_data_offset.deref(), picc_data_offset.deref() + 32);
    console.log(picc_data);

    var picc_data_str = picc_data.toString('utf-8');

    response.piccDataStr = picc_data_str

    var picc_data_tag = ref.alloc('byte', 0);
    
    var picc_data_buffer = Buffer.from(picc_data_str, 'hex')
    status = uFCoder.nt4h_decrypt_picc_data(picc_data_buffer, meta_data_key_pk, picc_data_tag, picc_uid, sdm_read_cnt)
    if (status) {
      response.Success = "Function status: Failed to decrypt PICC data."
      response.Status = uFR_NT4H_Status2String(status)
      response.UfrStatus = status
      //res.send(JSON.stringify(response));
      console.log(response)
  
      res.send(JSON.stringify(response));
      return;
    }


    var uid_str = ""
    for (i = 0; i < 7; i++) {
      uid_str += picc_uid[i].toString(16).padStart(2, '0').toUpperCase();
    }
    
    response.piccUidStr = uid_str;
    response.piccSDMReadingCounter = sdm_read_cnt.deref().toString();
  }

  if (meta_data_key_no.deref() == 0x0E) // PICC DATA NOT ENCRYPTED, meta data key == 14 #free access
  {

    if (uid_enable.deref()) {
      var picc_data = read_data.slice(uid_offset.deref(), uid_offset.deref() + 14);
      var picc_uid_str = picc_data.toString('utf-8');
      picc_uid = Buffer.from(picc_uid_str, 'hex');

      response.piccUidStr = picc_uid_str
    }

    if (read_ctr_enable.deref()) {
      var picc_data = read_data.slice(read_ctr_offset.deref(), read_ctr_offset.deref() + 6);
      var sdm_counter_str = picc_data.toString('utf-8');
      sdm_read_cnt = ref.alloc('int', Number.parseInt(sdm_counter_str, 16));
      response.piccSDMReadingCounter = sdm_read_cnt.deref().toString();
    }
  }

  if (enc_file_data_enable.deref())
  {
    // /nt4h_decrypt_sdm_enc_file_data
    var enc_data = read_data.slice(enc_offset.deref(), enc_offset.deref() + enc_length.deref());
    console.log(enc_data);

    var enc_data_str = enc_data.toString('utf-8');

    response.piccEncryptedFileData = enc_data_str

    var enc_data_buffer = Buffer.from(enc_data_str, 'hex')

    status = uFCoder.nt4h_decrypt_sdm_enc_file_data(sdm_read_cnt.deref(), picc_uid, file_data_read_key_pk, enc_data_buffer, (enc_length.deref() / 2))
    if (status) {
      response.Success = "Function status: Failed to decrypt enc. file data."
      response.Status = uFR_NT4H_Status2String(status)
      response.UfrStatus = status
      console.log(response)
      
      res.send(JSON.stringify(response));
      return;
    }

    var piccFileData_str = enc_data_buffer.toString('utf-8').replace(/\0/g, '')

    response.piccFileData = piccFileData_str
  }

  if(file_data_read_key_no.deref() <= 4)
	{
    //MAC exists
    //nt4h_check_sdm_mac

    var ascii_mac_data = read_data.slice(mac_offset.deref(), mac_offset.deref() + 16);

    var ascii_mac_data_str = ascii_mac_data.toString('utf-8');

    response.piccASCIIMACData = ascii_mac_data_str

    var ascii_mac_data_buffer = Buffer.from(ascii_mac_data_str, 'hex')

    var mac_in_len = ref.alloc('byte', mac_offset.deref() - mac_input_offset.deref()) 

    var ascii_mac_in_data = Buffer.alloc(256)
    if (mac_in_len.deref()) {
      ascii_mac_in_data = read_data.slice(mac_input_offset.deref(), mac_input_offset.deref() + mac_in_len.deref())
      var ascii_mac_in_data_str = ascii_mac_in_data.toString('utf-8');

      response.piccASCIIMACInputData = ascii_mac_in_data_str
    }

    //status = nt4h_check_sdm_mac(sdm_read_cnt, uid, file_data_aes_key, ascii_mac_in, mac_in_len, mac);
    status = uFCoder.nt4h_check_sdm_mac(sdm_read_cnt.deref(), picc_uid, file_data_read_key_pk, ascii_mac_in_data, mac_in_len.deref(), ascii_mac_data_buffer);
    if (status) {
      response.Success = "Function status: MAC is not correct."
      response.Status = uFR_NT4H_Status2String(status)
      response.UfrStatus = status
      console.log(response)
      res.send(JSON.stringify(response));
      return;
    }
  }

  response.ndefHexDataStr = read_data_str;
  response.ndefDataStr = ndef_context_dataStr;
  response.Success = "Function status: Data read successfully.";
  response.Status = uFR_NT4H_Status2String(status)

  console.log("/ReadSDM_POST() response: ");
  console.log(response);

  //var response = "";
  res.send(JSON.stringify(response));
});

///////////////////////////////////////////////////////////////////////

app.post("/writeSDMData", function (req, res) {

  console.log(req.body);

  
  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0
  }

  console.log(req.body)

  var global_file_no = ref.alloc('byte', 2);
  var global_key_no = ref.alloc('byte', 14);
  var global_comm_mode = ref.alloc('byte', 0);

  // NDEF Write auth. parameters
  var providedKey = Buffer.alloc(1, 0);
  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));

  var ndefAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));
  
  if (ndefAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  // SDM Change settings auth. parameters
  var sdmAuthMode = ref.alloc('byte', Number(req.body.authenticationMode_SDM))

  var current_change_key_no = ref.alloc('byte', req.body.current_change_key_no)

  var providedKey_SDM = Buffer.alloc(1, 0)
  var readerKeyIndex_SDM = ref.alloc('byte', Number(req.body.readerKeyIndex_SDM))

  if (sdmAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey_SDM = prepareAesKey(req.body.providedKeyStr_SDM)
  }

  var curr_comm_mode = ref.alloc('byte', 3);
  var new_comm_mode = ref.alloc('byte', 0)

  // NDEF File settings
  var read_key_no = ref.alloc('byte', req.body.read_key_no);
  var write_key_no = ref.alloc('byte', req.body.write_key_no);
  var read_write_key_no = ref.alloc('byte', req.body.read_write_key_no);
  var new_change_key_no = ref.alloc('byte', req.body.new_change_key_no);

  // SDM Settings
  var uid_enable = ref.alloc('byte', req.body.uid_enable);
  var read_ctr_enable = ref.alloc('byte', req.body.read_ctr_enable);
  var read_ctr_limit_enable = ref.alloc('byte', req.body.read_ctr_limit_enable);
  var enc_file_data_enable = ref.alloc('byte', Number(req.body.enc_file_data_enable));

  var meta_data_key_no = ref.alloc('byte', req.body.meta_data_key_no);
  var file_data_read_key_no = ref.alloc('byte', req.body.file_data_read_key_no);
  var read_ctr_key_no = ref.alloc('byte', req.body.read_ctr_key_no);

  // Offsets - need to be calculated
  
  var uid_offset = ref.alloc('int', 0);
  var read_ctr_offset = ref.alloc('int', 0);
  var picc_data_offset = ref.alloc('int', 0);
  var mac_input_offset = ref.alloc('int', 0);

  var enc_offset = ref.alloc('int', 0);
  var enc_length = ref.alloc('int', 0);
  var mac_offset = ref.alloc('int', 0);

  var read_ctr_limit = ref.alloc('int', Number(req.body.read_ctr_limit));

  var tt_status_enable = ref.alloc('byte', 0);
  var tt_status_offset = ref.alloc('int', 0);

  var is_picc_data_encrypted = Boolean(req.body.is_picc_data_encrypted)
  var sdm_mac_enabled = Boolean(req.body.sdm_mac_enabled)

  if (is_picc_data_encrypted == false)
  {
    meta_data_key_no = ref.alloc('byte', 0x0E);
    // picc data free access, not encrypted
  }

  if (sdm_mac_enabled == false)
  {
    file_data_read_key_no = ref.alloc('byte', 0x0F);
  }

  var file_data_for_encryption = req.body.file_data_for_encryption

  var status = 0x1001;

  // NDEF FILE CREATION
  var ndef_with_header = new Array(0x00, 0x00, 0xD1, 0x01, 0x00, 0x55, 0x00);
  var ndef_header_len = ndef_with_header.length
  
  
  var uri = req.body.uri //"https://www.d-logic.com/999?p=00000000000000000000000000000000&m=0000000000000000"
  console.log(uri)
  var ndef_buffer = strToHex(uri)
  for (i = 0; i < ndef_buffer.length; i++) {
    ndef_with_header.push(ndef_buffer[i])
  }


  ndef_with_header[ndef_with_header.length] = '?'.charCodeAt(0)

  var total_ndef_len = ndef_with_header.length

  // SET PICC DATA OFFSET(s)

  if (meta_data_key_no.deref() == 0x0E)
  {
    // picc data not encrypted 
      if (uid_enable.deref())
      {
        ndef_with_header.push('u'.charCodeAt(0));
        
        ndef_with_header.push('='.charCodeAt(0));
        
        for (i = 0; i < 14; i++) {
          ndef_with_header.push(0)
        }
        uid_offset = ref.alloc('int', total_ndef_len + 2)
        total_ndef_len += 16
        
      }

      if (read_ctr_enable.deref())
      {
        ndef_with_header.push('c'.charCodeAt(0));
        
        ndef_with_header.push('='.charCodeAt(0));
        
        for (i = 0; i < 6; i++) {
          ndef_with_header.push(0)
        }
        read_ctr_offset = ref.alloc('int', total_ndef_len + 2)
        total_ndef_len += 8
      }
  } else
  {
    // encrypt PICC data

    ndef_with_header.push('p'.charCodeAt(0));
        
    ndef_with_header.push('='.charCodeAt(0));
    
    for (i = 0; i < 32; i++) {
      ndef_with_header.push(0)
    }

    picc_data_offset = ref.alloc('int', total_ndef_len + 2)

    uid_offset = ref.alloc('int', 0)
    read_ctr_offset = ref.alloc('int', 0)
    total_ndef_len += 34

  }

  // File key 0x0F == MAC data does not exist.
  if (file_data_read_key_no.deref() != 0x0F)
  {
    // MACing offsets 

    if (enc_file_data_enable.deref())
    {
      ndef_with_header.push('e'.charCodeAt(0));
        
      ndef_with_header.push('='.charCodeAt(0));
      var enc_file_len = file_data_for_encryption.length
      var total_enc_file_len = enc_file_len
      if (total_enc_file_len % 16) 
      {
          total_enc_file_len += 16 - (enc_file_len % 16)
      }

      total_enc_file_len *= 2;

      var file_data_buffer = Buffer.from(file_data_for_encryption)
      
      for (i = 0; i < total_enc_file_len; i++) {
        ndef_with_header.push(0)
      }

      for (i = 0; i < enc_file_len; i++) {
        ndef_with_header[i + total_ndef_len + 2 ] = file_data_buffer[i];
      }
      
      enc_length = ref.alloc('int', total_enc_file_len)
      enc_offset = ref.alloc('int', total_ndef_len + 2)
      mac_input_offset = ref.alloc('int', total_ndef_len + 2)
      total_ndef_len += total_enc_file_len + 2
    } 

    // APPEND MAC
    ndef_with_header.push('m'.charCodeAt(0));
        
    ndef_with_header.push('='.charCodeAt(0));

    for (i = 0; i < 16; i++) {
      ndef_with_header.push(0)
    }

    mac_offset = ref.alloc('int', total_ndef_len + 2)
    if (enc_file_data_enable.deref() == false)
    {
      mac_input_offset = ref.alloc('int', total_ndef_len + 2)
    }
    total_ndef_len += 18

  }

  ndef_with_header[1] = (total_ndef_len - ndef_header_len) + 5;
  ndef_with_header[4] = (total_ndef_len - ndef_header_len) + 1;

  // Get card type

  var dl_card_type = ref.alloc('byte', 0);
  status = uFCoder.GetDlogicCardType(dl_card_type);
  if (status) {
    response.Success = "Function status: Geting card type failed."
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status;

    console.log(response)

    res.send(JSON.stringify(response));
    return;
  }

  // WRITE NDEF MESSAGE PAYLOAD

  var global_file_no = ref.alloc('byte', 2);
  var global_key_no = ref.alloc('byte', write_key_no.deref()); // ndef write key set
  var global_comm_mode = ref.alloc('byte', 0); // set to plain for ndef write

  status = uFCoder.nt4h_set_global_parameters(global_file_no.deref(), global_key_no.deref(), global_comm_mode.deref());
  if (status) {
    response.Success = "Function status: Global parameters setting failed"
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status;

    console.log(response)
    res.send(JSON.stringify(response));
    return;
  }

  var payload_buffer = Buffer.from(ndef_with_header)
  var bytes_ret = ref.alloc('int', 0)
  
  if (ndefAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH)
  {
    status = uFCoder.LinearWrite_PK(payload_buffer, 0, total_ndef_len, bytes_ret, ndefAuthMode.deref(), providedKey_SDM)

  } else
  {
    status = uFCoder.LinearWrite(payload_buffer, 0, total_ndef_len, bytes_ret, ndefAuthMode.deref(), readerKeyIndex_SDM.deref())

  }

  if (status) {
    response.Success = "Function status: Failed to write NDEF message."
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
    console.log(response)
    res.send(JSON.stringify(response));
    return;
  }
  

    // CHANGE SDM FILE SETTINGS

  if (sdmAuthMode.deref() == T4T_AUTHENTICATION.T4T_RKA_PWD_AUTH) {
    // reader key authentication
    if (dl_card_type.deref() == 0x13) // if DL_NTAG_424_DNA_TT
    {
      status = uFCoder.nt4h_tt_change_sdm_file_settings(readerKeyIndex_SDM.deref(), global_file_no.deref(), current_change_key_no.deref(),
        curr_comm_mode.deref(), new_comm_mode.deref(),
        read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
        uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
        meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
        uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
        enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref(),
        tt_status_enable.deref(), tt_status_offset.deref())

    } else {
      //status = uFCoder.nt4h_change_sdm_file_settings()
      status = uFCoder.nt4h_change_sdm_file_settings(readerKeyIndex.deref(), file_no.deref(), curr_change_key_no.deref(),
        curr_comm_mode.deref(), new_comm_mode.deref(),
        read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
        uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
        meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
        uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
        enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref())
    }
  } else {
    if (dl_card_type.deref() == 0x13) // if DL_NTAG_424_DNA_TT
    {
      status = uFCoder.nt4h_tt_change_sdm_file_settings_pk(providedKey, file_no.deref(), curr_change_key_no.deref(),
        curr_comm_mode.deref(), new_comm_mode.deref(),
        read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
        uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
        meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
        uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
        enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref(),
        tt_status_enable.deref(), tt_status_offset.deref())
    } else {
      status = uFCoder.nt4h_change_sdm_file_settings_pk(providedKey, file_no.deref(), curr_change_key_no.deref(),
        curr_comm_mode.deref(), new_comm_mode.deref(),
        read_key_no.deref(), write_key_no.deref(), read_write_key_no.deref(), new_change_key_no.deref(),
        uid_enable.deref(), read_ctr_enable.deref(), read_ctr_limit_enable.deref(), enc_file_data_enable.deref(),
        meta_data_key_no.deref(), file_data_read_key_no.deref(), read_ctr_key_no.deref(),
        uid_offset.deref(), read_ctr_offset.deref(), picc_data_offset.deref(), mac_input_offset.deref(),
        enc_offset.deref(), enc_length.deref(), mac_offset.deref(), read_ctr_limit.deref())
    }
  }

  if (status) {
    response.Success = "Function status: Failed to write SDM file settings.";
  } else {
    response.Success = "Function status: Card programming successful"
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

app.post("/linearRead", function (req, res) {

  
  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0,

    "ReadData": ""
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var file_no = ref.alloc('byte', Number(req.body.file_no))
  var read_key_no = ref.alloc('byte', req.body.read_key_no);
  var comm_mode = ref.alloc('byte', req.body.communication_mode); // set to plain for ndef write

  var linear_address = ref.alloc('int', req.body.linear_address)
  var linear_length = ref.alloc('int', req.body.linear_length)
  
  var read_buffer = Buffer.alloc(linear_length.deref())
  var bytes_ret = ref.alloc('int', 0)

  var status = 0x1001;

  status = uFCoder.nt4h_set_global_parameters(file_no.deref(), read_key_no.deref(), comm_mode.deref());
  if (status) {
    response.Success = "Function status: Global parameters setting failed"
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status;

    console.log(response)
    res.send(JSON.stringify(response));
    return;
  }

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) 
  {
    status = uFCoder.LinearRead_PK(read_buffer, linear_address.deref(), linear_length.deref(), bytes_ret, readerAuthMode.deref(), providedKey)
  } else 
  {
    status = uFCoder.LinearRead(read_buffer, linear_address.deref(), linear_length.deref(), bytes_ret, readerAuthMode.deref(), readerKeyIndex.deref())
  }

  if (status) {
    response.Success = "Function status: Linear reading has failed.";
    response.ReadData = ""
  } else {
    response.Success = "Function status: Linear reading was successful."

    var output = req.body.output;

    if (output == 0) {
      // ASCII STR
      response.ReadData = read_buffer.toString('utf-8');
    } else {
      // HEX STR
      var data_str = ""
      for (i = 0; i < read_buffer.length; i++) {
        data_str += read_buffer[i].toString(16).padStart(2, '0').toUpperCase() + ":";
      }
      response.ReadData = data_str.slice(0, -1)
    }
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/linearWrite", function (req, res) {

  
  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var file_no = ref.alloc('byte', Number(req.body.file_no))
  var write_key_no = ref.alloc('byte', req.body.write_key_no);
  var comm_mode = ref.alloc('byte', req.body.communication_mode); // set to plain for ndef write

  var linear_address = ref.alloc('int', req.body.linear_address)
  var linear_length = ref.alloc('int', req.body.linear_length)

  var data_str = req.body.data
  
  var status = 0x1001;

  status = uFCoder.nt4h_set_global_parameters(file_no.deref(), write_key_no.deref(), comm_mode.deref());
  if (status) {
    response.Success = "Function status: Global parameters setting failed"
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status;

    console.log(response)
    res.send(JSON.stringify(response));
    return;
  }

  var input = req.body.input

  if (input == 0)
  {
    var write_buffer = Buffer.from(data_str)
    // ASCI BUFFER
  } else 
  {
    // WRITE HEX
    data_str = data_str.replace(/:/g, "")
    var write_buffer = Buffer.from(data_str, 'hex')
  }

  var bytes_ret = ref.alloc('int', 0)
  
  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH)
  {
    status = uFCoder.LinearWrite_PK(write_buffer, linear_address.deref(), linear_length.deref(), bytes_ret, readerAuthMode.deref(), providedKey)

  } else
  {
    status = uFCoder.LinearWrite(write_buffer, linear_address.deref(), linear_length.deref(), bytes_ret, readerAuthMode.deref(), readerKeyIndex.deref())

  }

  if (status) {
    response.Success = "Function status: Linear writing has failed.";
    response.ReadData = ""
  } else {
    response.Success = "Function status: Linear writing was successful."
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/getUid", function (req, res) {

  
  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0,
    "Uid": ""
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var status = 0x1001;

  var key_no = ref.alloc('byte', req.body.key_no)
  var uid = Buffer.alloc(7);

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    status = uFCoder.nt4h_get_uid_pk(providedKey, key_no.deref(), uid);
  } else {
    status = uFCoder.nt4h_get_uid(readerKeyIndex.deref(), key_no.deref(), uid);
  }

  if (status) {
    response.Success = "Function status: Get UID has failed.";
    
  } else {
    response.Success = "Function status: Get UID was successful."

    var uid_str = ""
    for (i = 0; i < 7; i++) {
      uid_str += uid[i].toString(16).padStart(2, '0').toUpperCase() +":";
    }
    response.Uid = uid_str.slice(0, -1)
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/setRandomUid", function (req, res) {

  
  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var status = 0x1001;


  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    status = uFCoder.nt4h_set_rid_pk(providedKey);
  } else {
    status = uFCoder.nt4h_set_rid(readerKeyIndex.deref());
  }

  if (status) {
    response.Success = "Function status: Set Random UID has failed.";
    
  } else {
    response.Success = "Function status: Set Random UID was successful."
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/getSdmCounter", function (req, res) {

  
  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0,
    "SdmCounter": ""
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var file_no = ref.alloc('byte', Number(req.body.file_no))
  var key_no = ref.alloc('byte', req.body.key_no);

  var status = 0x1001;

  var sdm_read_ctr = ref.alloc('int', 0)

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    status = uFCoder.nt4h_get_sdm_ctr_pk(providedKey, file_no.deref(), key_no.deref(), sdm_read_ctr);
  } else if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_RKA_PWD_AUTH) {
    status = uFCoder.nt4h_get_sdm_ctr(readerKeyIndex.deref(), file_no.deref(), key_no.deref(), sdm_read_ctr);
  } else {
    status = uFCoder.nt4h_get_sdm_ctr_no_auth(file_no.deref(), sdm_read_ctr);
  }

  if (status) {
    response.Success = "Function status: Get SDM Counter has failed.";
    
  } else {
    response.Success = "Function status: Get SDM Counter was successful."
    response.SdmCounter = sdm_read_ctr.deref().toString()
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/tagTamperEnable", function (req, res) {

  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var tt_status_key_no = ref.alloc('byte', req.body.key_no);

  var status = 0x1001;

  var sdm_read_ctr = ref.alloc('int', 0)

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    status = uFCoder.nt4h_enable_tt_pk(providedKey, tt_status_key_no.deref());
  } else {
    status = uFCoder.nt4h_enable_tt(readerKeyIndex.deref(), file_no.deref(), tt_status_key_no.deref(), sdm_read_ctr);
  }

  if (status) {
    response.Success = "Function status: TT Status Enable has failed.";
    
  } else {
    response.Success = "Function status: TT Status Enable was successful."
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/getTagTamperStatus", function (req, res) {

  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0,
    "TTPermStatus": "",
    "TTCurrStatus": "",
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var tt_status_key_no = ref.alloc('byte', req.body.key_no);

  var status = 0x1001;

  var tt_perm_status = ref.alloc('byte', 0)
  var tt_curr_status = ref.alloc('byte', 0)

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    status = uFCoder.nt4h_get_tt_status_pk(providedKey, tt_status_key_no.deref(), tt_perm_status, tt_curr_status);
  } else if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_RKA_PWD_AUTH){
    status = uFCoder.nt4h_get_tt_status(readerKeyIndex.deref(), tt_status_key_no.deref(), tt_perm_status, tt_curr_status);
  } else {
    status = uFCoder.nt4h_get_tt_status_no_auth(tt_perm_status, tt_curr_status);
  }

  if (status) {
    response.Success = "Function status: Get TT Status has failed.";  
  } else {
    response.Success = "Function status: Get TT Status was successful."
    response.TTPermStatus = tt_perm_status.toString('utf-8')
    response.TTCurrStatus = tt_curr_status.toString('utf-8')
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/originalityCheck", function (req, res) {

  console.log(req.body);

  var response = {
    "Status": "",
    "Success": "",
    "UfrStatus": 0,
    "Signature": "",
    "Result": "",
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var tt_status_key_no = ref.alloc('byte', req.body.key_no)

  var status = 0x1001;

  var uid = new Buffer.alloc(11);
  var sak = ref.alloc('byte');
  var uid_len = ref.alloc('byte');
  
  status = uFCoder.GetCardIdEx(sak, uid, uid_len);
  if (status)
  {
    response.Success = "Function status: Geting card UID failed"
    response.Status = uFR_NT4H_Status2String(status)
    response.UfrStatus = status
    res.send(JSON.stringify(response));
    return;
  }

  var ecc_signature = Buffer.alloc(255)
  var ecc_sig_len = ref.alloc('byte', 0)
  var dl_card_type = ref.alloc('byte', 0)

  if (uid_len.deref() == 4)
  {
    // Random UID detected, authentication required
    if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH)
    {
      // PK
      status = uFCoder.nt4h_rid_read_ecc_signature_pk(providedKey, tt_status_key_no.deref(), uid, ecc_signature, dl_card_type);
    
    } else
    {
      status = uFCoder.nt4h_rid_read_ecc_signature(readerKeyIndex.deref(), tt_status_key_no.deref(), uid, ecc_signature, dl_card_type);
    
    }
  } else 
  {
    status = uFCoder.ReadECCSignatureExt(ecc_signature, ecc_sig_len, uid, uid_len, dl_card_type);
  }

  if (status) {
    response.Success = "Function status: ECC signature read has failed.";  
  } else {
    response.Success = "Function status: ECC signature read was successful."

    var signature_str = ""
    for (i = 0; i < ecc_sig_len.deref(); i++) {
      signature_str += ecc_signature[i].toString(16).padStart(2, '0').toUpperCase() + ":";
    }
    response.ReadData = signature_str.slice(0, -1)
    response.Signature = signature_str

    status = uFCoder.OriginalityCheck(ecc_signature, uid, uid_len.deref(), dl_card_type.deref())
    if (status == 0) //UFR_OK
    {
      response.Result = "Function status: Tag is NXP genuine"
    } else if (status == 0x200) //UFR_NOT_NXP_GENUINE
    {
      response.Result = "Function status: Tag is NOT NXP genuine"
    } else if (status == 0x201) //UFR_OPEN_SSL_DYNAMIC_LIB_FAILED
    {
      response.Result = "Function status: OpenSSL library error"
    } else if (status == 0x202) //UFR_OPEN_SSL_DYNAMIC_LIB_NOT_FOUND
    {
      response.Result = "Function status: OpenSSL library not found"
    } else
    {
      response.Result = "Function status: Originality check encountered an error."
    }
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/changeCardKey", function (req, res) {

  console.log(req.body);

  var response = {
    "Status": "",
    "Success": ""
  }

  var readerKeyIndex = ref.alloc('byte', Number(req.body.readerKeyIndex));
  var readerAuthMode = ref.alloc('byte', Number(req.body.authenticationMode));

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH) {
    providedKey = prepareAesKey(req.body.providedKeyStr)
  }

  var key_no = ref.alloc('byte', req.body.key_no)

  var old_key = prepareAesKey(req.body.old_key)
  var new_key = prepareAesKey(req.body.new_key)

  var status = 0x1001;

  if (readerAuthMode.deref() == T4T_AUTHENTICATION.T4T_PK_PWD_AUTH)
  {
    // PK change
    status = uFCoder.nt4h_change_key_pk(providedKey, key_no.deref(), new_key, old_key);
  } else 
  {
    //RK Change
    status = uFCoder.nt4h_change_key(readerKeyIndex.deref(), key_no.deref(), new_key, old_key);
  }

  if (status)
  {
    response.Success = "Function status: Change Card key #" +key_no.deref().toString()+" has failed";
  } else 
  {
    response.Success = "Function status: Change Card key #" +key_no.deref().toString()+" was successful";
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

app.post("/storeReaderKey", function (req, res) {

  console.log(req.body);

  var response = {
    "Status": "",
    "Success": ""
  }

  var key_no = ref.alloc('byte', req.body.key_no)
  
  var aes_key = prepareAesKey(req.body.aes_key)

  var status = 0x1001;

  status = uFCoder.uFR_int_DesfireWriteAesKey(key_no.deref(), aes_key);

  if (status)
  {
    response.Success = "Function status: Store Key #" + key_no.deref()+" has failed";
  } else 
  {
    response.Success = "Function status: Store Key #" + key_no.deref()+" was successful";  
  }

  response.Status = uFR_NT4H_Status2String(status)
  response.UfrStatus = status;

  console.log(response)

  res.send(JSON.stringify(response))

});

///////////////////////////////////////////////////////////////////////

function delay(time) {
  return new Promise(resolve => setTimeout(resolve, time));
}

////////////////////////////////////////////////////////////////////////////////////////

app.listen(port, hostname, function (err) {
  if (err)
    return console.log(err);
  console.log("Listening at http://%s:%s", hostname, port);
});

////////////////////////////////////////////////////////////////////////////////////////
// Helper Functions
////////////////////////////////////////////////////////////////////////////////////////

// Convert a hex string to a byte array
function strToHex(str) {
  var result = [];
  for (var i = 0; i < str.length; i++) {
    result.push(str.charCodeAt(i));
  }
  return result;
}

function hexToBytes(hex) {
  if (hex.includes(":")) {
    hex = hex.replace(/:/g, "")
  }

  if (hex.includes(" ")) {
    hex = hex.replace(/ /g, "")
  }

  let bytes = [];
  for (let c = 0; c < hex.length; c += 2)
    bytes.push((hex.substr(c, 2), 16));
  return bytes;
}

function prepareAesKey(keyStr) {

  if (keyStr.includes(":")) {
    keyStr = keyStr.replace(/:/g, "")
  }

  if (keyStr.length < 32) {
    return Buffer.alloc(1, 0)
  }
  return Buffer.from(keyStr, 'hex')

}

function uFR_NT4H_Status2String(status) {
  if ((status < 0xC0) || (status > 0xCB)) {
    return uFCoder.UFR_Status2String(status)
  } else {
    var nt4h_status = "[0x" + status.toString(16).toUpperCase() + " (" + status.toString() + ")] "
    switch (status) {
      case 0xC0:
        nt4h_status += 'NT4H_COMMAND_ABORTED'
        break;
      case 0xC1:
        nt4h_status += 'NT4H_LENGTH_ERROR'
        break;
      case 0xC2:
        nt4h_status += 'NT4H_PARAMETER_ERROR'
        break;
      case 0xC3:
        nt4h_status += 'NT4H_NO_SUCH_KEY'
        break;
      case 0xC4:
        nt4h_status += 'NT4H_PERMISSION_DENIED'
        break;
      case 0xC5:
        nt4h_status += 'NT4H_AUTHENTICATION_DELAY'
        break;
      case 0xC6:
        nt4h_status += 'NT4H_MEMORY_ERROR'
        break;
      case 0xC7:
        nt4h_status += 'NT4H_INTEGRITY_ERROR'
        break;
      case 0xC8:
        nt4h_status += 'NT4H_FILE_NOT_FOUND'
        break;
      case 0xC9:
        nt4h_status += 'NT4H_BOUNDARY_ERROR'
        break;
      case 0xCA:
        nt4h_status += 'NT4H_INVALID_MAC'
        break;
      case 0xCB:
        nt4h_status += 'NT4H_NO_CHANGES'
        break;
      default:
        break;
    }
    return nt4h_status;
  }
}
